/* fsgrab.c -- get blocks from an (ext2) filesystem on Linux
 *
 * Aaron Crane <aaronc@pobox.com>
 *
 * `fsgrab', a program which reads blocks from an unmounted filesystem, is
 * Copyright  1997, 1998 Aaron Crane <aaronc@pobox.com>.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 *
 * 1997-10-17: v1.1
 *
 *   Improved program structure.
 *   Now reads until error or eof or found enough data.
 *   Now writes until error or written enough data.
 *   Added `-f' (force stdout writes) option.
 *
 * 1998-09-10: v1.2
 *
 *   Autoconf-ized.
 *   Also works on systems without gcc, gmake, glibc
 *   Attempt to use ext2fs_llseek or llseek where available.
 */

#if HAVE_CONFIG_H
#include <fsg-conf.h>
#endif

#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>

static const char *invocation_name = "";

#ifndef __GNUC__
#define __attribute__(attrs)
#endif

static void message (const char *fmt, ...) __attribute__ ((format (printf, 1, 2)));
static void fatal (const char *fmt, ...) __attribute__ ((format (printf, 1, 2), noreturn));
static void *mustmalloc (size_t n);

static int fsgrab (const char *dev, size_t blocksize, size_t count, size_t skip);
static int do_open (const char *dev);
static void do_seek (const char *dev, int fd, size_t blocksize, size_t count);
static ssize_t do_read (const char *dev, int fd, unsigned char *buf, ssize_t n);
static void do_write (const unsigned char *buf, ssize_t n);

static void
usage (int status)
{
    fprintf (stderr,
	     "Usage: %s [-b BLOCKSIZE] [-c COUNT] [-s SKIP] [-f] [DEVICE]\n"
	     "       %s -h\n"
	     "-b: BLOCKSIZE is the size of a filesystem block; defaults to 1024\n"
	     "-c: COUNT is count of blocks to read; defaults to 1\n"
	     "-s: SKIP is count of blocks to skip; defaults to 0\n"
	     "-f: Force writing to stdout, even if it's a terminal\n"
	     "-h: print this help message\n"
	     "DEVICE is the device to read from; defaults to /dev/hda1\n",
	     invocation_name, invocation_name);
    exit (status);
}

int
main (int argc, char **argv)
{
    int opt;
    const char *device = "/dev/hda1";
    size_t blocksize = 1024, count = 1, skip = 0;
    int avoid_ttys = 1;

    invocation_name = strrchr (argv[0], '/');
    if (invocation_name)
	invocation_name++;
    else
	invocation_name = argv[0];

    while ((opt = getopt (argc, argv, "b:c:s:fh")) != EOF)
	switch (opt)
	{
	case 'b':
	    blocksize = strtoul (optarg, 0, 0);
	    break;
	case 'c':
	    count = strtoul (optarg, 0, 0);
	    break;
	case 's':
	    skip = strtoul (optarg, 0, 0);
	    break;
	case 'f':
	    avoid_ttys = 0;
	    break;
	case 'h':
	    usage (0);
	default:
	    usage (1);
	}

    if (optind < argc)
	device = argv[optind];

    if (avoid_ttys && isatty (1))
	    fatal ("%s: stdout is a terminal\n", ttyname (1));

    return fsgrab (device, blocksize, count, skip) == -1;
}

static int
fsgrab (const char *device, size_t blocksize, size_t count, size_t skip)
{
    int fd;
    unsigned char *buf;
    ssize_t n;

    fd = do_open (device);
    do_seek (device, fd, skip, blocksize);
    buf = mustmalloc (count * blocksize);
    n = do_read (device, fd, buf, count * blocksize);
    do_write (buf, n);
    free (buf);
    return 0;
}

static int
do_open (const char *device)
{
    int fd;

    fd = open (device, O_RDONLY);
    if (fd < 0)
	fatal ("couldn't open device %s: %s\n", device, strerror (errno));
    return fd;
}


static ssize_t
do_read (const char *device, int fd, unsigned char *buf, ssize_t n)
{
    ssize_t i;
    unsigned char *p = buf;

    while (n > 0)
    {
	i = read (fd, p, n);
	if (i == -1)
	{
	    message ("%s: read failed after %lu chars: %s\n",
		     device, (unsigned long) (p - buf), strerror (errno));
	    /* last-ditch attempt to write what we've got */
	    do_write (buf, p - buf);
	    exit (1);
	}
	if (i == 0)
	{
	    message ("%s: warning: unexpected end of file after %lu chars\n",
		     device, (unsigned long) (p - buf));
	    break;
	}
	n -= i;
	p += i;
    }

    return p - buf;
}

static void
do_write (const unsigned char *buf, ssize_t n)
{
    ssize_t i;
    const unsigned char *p = buf;

    while (n > 0)
    {
	i = write (1, p, n);
	if (i == -1)
	    fatal ("write on stdout failed: %s\n", strerror (errno));
	n -= i;
	p += i;
    }
}

#if HAVE_LIBEXT2FS

#if HAVE_EXT2FS_EXT2FS_H

/* <ext2fs/ext2fs.h> uses these structs without even a forward declaration
 * for them, which causes warnings for me because I don't include
 * <linux/ext2_fs.h>. */
struct ext2_inode; struct ext2_dir_entry;

/* <ext2fs/ext2fs.h> has a bunch of inline functions which dereference
 * pointers to structs which aren't defined if you don't include kernel
 * files.  It lets you define this bletcherous token to disable that
 * breakage.
 */
#define NO_INLINE_FUNCS

#include <ext2fs/ext2fs.h>
#endif

#define SEEK(fd, offset, whence) ext2fs_llseek (fd, offset, whence)
#define SEEKNAME "ext2fs_llseek"
typedef ext2_loff_t my_offset_t;

#elif HAVE_LLSEEK
#define SEEK(fd, offset, whence) llseek (fd, offset, whence)
#define SEEKNAME "llseek"
typedef loff_t my_offset_t;

#else /* Neither ext2fs_llseek nor llseek */
#define SEEK(fd, offset, whence) lseek (fd, offset, whence)
#define SEEKNAME "lseek"
typedef off_t my_offset_t;
#endif

static void
do_seek (const char *device, int fd, size_t blocksize, size_t count)
{
    my_offset_t seekpos = (my_offset_t) blocksize * count;

    if (seekpos > 0)
    {
	my_offset_t i = SEEK (fd, seekpos, SEEK_SET);
	if (i == -1 || i != seekpos)
	    fatal ("%s: " SEEKNAME " failed: %s\n", device, strerror (errno));
    }
}


/*
 * `Library' routines
 */

static void *
mustmalloc (size_t n)
{
    void *p = malloc (n);

    if (!p)
	fatal ("out of memory\n");
    return p;
}

inline static void
vmessage (const char *fmt, va_list va)
{
    fprintf (stderr, "%s: ", invocation_name);
    vfprintf (stderr, fmt, va);
}

static void
message (const char *fmt, ...)
{
    va_list va;

    va_start (va, fmt);
    vmessage (fmt, va);
    va_end (va);
}

static void
fatal (const char *fmt, ...)
{
    va_list va;

    va_start (va, fmt);
    vmessage (fmt, va);
    va_end (va);
    exit (1);
}
